'''EfficientNet in PyTorch.

Paper: "EfficientNet: Rethinking Model Scaling for Convolutional Neural Networks".
'''
import torch
import torch.nn as nn
import torch.nn.functional as F


from .build import META_ARCH_REGISTRY

__all__ = [
    'EfficientNet',
    'EfficientNetB0',
]

class Block(nn.Module):
    '''expand + depthwise + pointwise + squeeze-excitation'''

    def __init__(self, in_planes, out_planes, expansion, stride):
        super(Block, self).__init__()
        self.stride = stride

        planes = expansion * in_planes
        self.conv1 = nn.Conv2d(
            in_planes, planes, kernel_size=1, stride=1, padding=0, bias=False)
        self.bn1 = nn.BatchNorm2d(planes)
        self.conv2 = nn.Conv2d(planes, planes, kernel_size=3,
                               stride=stride, padding=1, groups=planes, bias=False)
        self.bn2 = nn.BatchNorm2d(planes)
        self.conv3 = nn.Conv2d(
            planes, out_planes, kernel_size=1, stride=1, padding=0, bias=False)
        self.bn3 = nn.BatchNorm2d(out_planes)

        self.shortcut = nn.Sequential()
        if stride == 1 and in_planes != out_planes:
            self.shortcut = nn.Sequential(
                nn.Conv2d(in_planes, out_planes, kernel_size=1,
                          stride=1, padding=0, bias=False),
                nn.BatchNorm2d(out_planes),
            )

        # SE layers
        self.fc1 = nn.Conv2d(out_planes, out_planes//16, kernel_size=1)
        self.fc2 = nn.Conv2d(out_planes//16, out_planes, kernel_size=1)

    def forward(self, x):
        out = F.relu(self.bn1(self.conv1(x)))
        out = F.relu(self.bn2(self.conv2(out)))
        out = self.bn3(self.conv3(out))
        shortcut = self.shortcut(x) if self.stride == 1 else out
        # Squeeze-Excitation
        w = F.avg_pool2d(out, out.size(2))
        w = F.relu(self.fc1(w))
        w = self.fc2(w).sigmoid()
        out = out * w + shortcut
        return out

@META_ARCH_REGISTRY.register()
class EfficientNet(nn.Module):
    def __init__(self, layers, num_classes=10):
        super(EfficientNet, self).__init__()
        self.layers = layers
        self.conv1 = nn.Conv2d(3, 32, kernel_size=3,
                               stride=1, padding=1, bias=False)
        self.bn1 = nn.BatchNorm2d(32)
        self.layers = self._make_layers(in_planes=32)
        self.last_linear = nn.Linear(layers[-1][1], num_classes)

    def _make_layers(self, in_planes):
        layers = []
        for expansion, out_planes, num_blocks, stride in self.layers:
            strides = [stride] + [1]*(num_blocks-1)
            for stride in strides:
                layers.append(Block(in_planes, out_planes, expansion, stride))
                in_planes = out_planes
        return nn.Sequential(*layers)

    def features(self, input): # 2*3*32*32
        x = F.relu(self.bn1(self.conv1(input))) # 2*32*32*32
        x = self.layers(x) # 2*320*1*1
        return x

    def logits(self, features):
        x = nn.AdaptiveAvgPool2d((1,1))(features) # 2*320*1*1
        x = x.view(x.size(0), -1)
        x = self.last_linear(x)
        return x

    def forward(self, input): # 2*3*32*32
        x = self.features(input) # 2*320*1*1
        x = self.logits(x) # 2*num_classes
        return x


@META_ARCH_REGISTRY.register()
class EfficientNetB0(EfficientNet):
    def __init__(self, cfg):
        num_classes = cfg.model.classes
        # (expansion, out_planes, num_blocks, stride)
        layers = [(1,  16, 1, 2),
            (6,  24, 2, 1),
            (6,  40, 2, 2),
            (6,  80, 3, 2),
            (6, 112, 3, 1),
            (6, 192, 4, 2),
            (6, 320, 1, 2)]
        super(EfficientNetB0, self).__init__(layers, num_classes)